%% Gesichtserkennung mit Eigenvektoren oder (besser) Singulrvektoren
clear variables;
close all;
load Yale_64x64;
% Datenquelle: http://www.cad.zju.edu.cn/home/dengcai/Data/FaceData.html
% Yale database
% Contains 165 grayscale images in GIF format of 15 individuals. There are
% 11 images per subject, one per different facial expression or
% configuration: center-light, w/glasses, happy, left-light, w/no glasses,
% normal, right-light, sad, sleepy, surprised, and wink.
% 64x64 Data File: contains variables 'fea' and 'gnd'. 
% Each row of 'fea' is a face; 'gnd' is the label. 

%% einige zufllige Bilder aus dieser Datenbank
% Achtung: Bilder entsprechen Zeilen in fea!
% fea hat 165 Zeilen, jede Zeile hat 4096 Werte, diese in einem
% 64x64-Quadrat angeordnet ergeben ein Bild. Der r-Vektor pickt 16 Zeilen
% zufllig heraus.
r = randi([1  165],1,16);

figure,subplot(4,4,1)
colormap gray
for i = 1:16
    subplot(4,4,i )
    imagesc(reshape(fea(r(i),:),[64,64]))
    axis off
end


%%
%subtract mean
% der mean-Befehl berechnet die Mittelwerte der Spalten von fea, ergibt
% also einen 4096-Vektor; der entspricht einem 64x64-Quadrat, deswegen ist
% hier von mean-matrix die Rede. Dies ist das "Mittelgesicht" aller 165
% Bilder in fea.
mean_matrix = mean(fea);
%x = bsxfun(@minus, fea, mean_matrix); % in aelteren Matlab Versionen
x = fea - mean_matrix;

%% Original-Algorithmus
% erzeugt hier Riesen-Matrix und Riesen-Eigenvektor-Zerlegung
% (so hat man Anfang der 1990er Jahre gerechnet; heute rechnet man anders)
XTX = x'*x;
[V,D] = eig(XTX);
eigval = diag(D);
% sort eigenvalues and eigenvectors in descending order
eigval = eigval(end:-1:1);
V = fliplr(V);

%% show mean and 1th through 15th principal eigenvectors
figure,subplot(4,4,1)
imagesc(reshape(mean_matrix, [64,64]))
axis off
colormap gray
for i = 1:15
    subplot(4,4,i+1)
    imagesc(reshape(V(:,i),[64,64]))
    axis off
end

%% Algorithmus mit SVD
% viel effizienter, weil die gro"se XTX-Matrix nicht berechnet wird
[U,S,V] = svd(x);
%V = fliplr(V);
% berechnet dieselben Eigenfaces, es kann nur sein, dass das Vorzeichen
% anders ist
figure,subplot(4,4,1)
imagesc(fliplr(reshape(mean_matrix, [64,64])))
axis off
colormap gray
for i = 1:15
    subplot(4,4,i+1)
    imagesc(reshape(V(:,i),[64,64]))
    axis off
end
 


%% evaluate the number of principal components needed to represent 95% Total variance.
eigsum = sum(eigval);
csum = 0;
for i = 1:64*64
    csum = csum + eigval(i);
    tv = csum/eigsum;
    if tv > 0.95
        k95 = i; % break from loop when 95% reached
        break
    end
end

% Dieser Plot zeigt den Grad der Approximation abhaengig von den
% Eigengesichtern.
 
plot(cumsum(eigval(1:i))/eigsum)

%%
% Jedes Bild aus der Datenbank lsst sich aus Eigengesichtern
% rekonstruieren:

% Ein Originalbild
imagesc(reshape(x(17,:),[64,64]))
colormap gray
axis off
axis equal

% Diese 45 Werte geben das Mischungsverhltnis der Eigenbilder an; 
comp=x(17,:)*V(:,1:45); 


%% Rekonstruktion aus Mittelgesicht plus comp*Eigengesichter 
figure
reconstr=mean_matrix+comp(1:15)*V(:,1:15)'; % aus den ersten 15
imagesc(reshape(reconstr,[64,64]))
colormap gray
axis off
axis equal
figure
reconstr=mean_matrix+comp(1:30)*V(:,1:30)'; % aus den ersten 30
imagesc(reshape(reconstr,[64,64]))
colormap gray
axis off
axis equal

%figure,subplot(4,4,1)

